#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <pthread.h>
#include <gtk/gtk.h>
#include <linux/types.h> 
#include <linux/videodev.h>
#include <stdio.h>
#include <jpeglib.h>
#include "gqcam.h"
#include "frontend.h"
#include "save.h"




struct Page_Ppm page_ppm;
struct Page_Jpeg page_jpeg;
struct Page_Gif page_gif;
GtkWidget *textbox;


void select_ok(GtkWidget *widget, GdkEvent *event, struct Camera *camera)
{
  gtk_entry_set_text(GTK_ENTRY(textbox), gtk_file_selection_get_filename(GTK_FILE_SELECTION (widget)));
  gtk_widget_destroy(GTK_WIDGET(widget));
  return;
}

void select_cancel(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  gtk_widget_destroy(GTK_WIDGET(widget));
  return;
}

void file_select(GtkWidget *widget, struct Camera *camera)
{
  GtkWidget *fileselect;

  //  camera->saving = 1;
  fileselect = gtk_file_selection_new("Save Image");
  gtk_widget_show(fileselect);

  gtk_signal_connect_object (GTK_OBJECT (GTK_FILE_SELECTION (fileselect)->cancel_button), "clicked", (GtkSignalFunc) select_cancel, GTK_OBJECT (fileselect));

  gtk_signal_connect_object (GTK_OBJECT (GTK_FILE_SELECTION (fileselect)->ok_button), "clicked", (GtkSignalFunc) select_ok, GTK_OBJECT (fileselect));

  return;
}


void save_ok(GtkWidget *widget, struct Camera *camera)
{
  int ls=0, x, y;
  int i;
  int lines, pixels_per_line;
  FILE *outfile;
  unsigned char buff[3];

  struct jpeg_compress_struct cinfo;
  struct jpeg_error_mgr jerr;
  JSAMPROW row_pointer[1];     // pointer to JSAMPLE row(s)
  int row_stride;              // physical row width in image buffer

  if(!(outfile = fopen(gtk_entry_get_text(GTK_ENTRY(textbox)), "w"))){
    printf("Need a file name...\n");
    return;
  }
  
  switch (camera->savetype) {

  case PPM:
    if (GTK_TOGGLE_BUTTON(page_ppm.radio_raw_ppm)->active){
      fprintf (outfile, "P6\n%d %d\n%d\n", camera->vid_win.width, camera->vid_win.height, 255);
      if (camera->greyscale){
	for (i = 0; i < camera->vid_win.width*camera->vid_win.height; i++){
	  buff[0] = camera->savebuf[i];
	  buff[1] = camera->savebuf[i];
	  buff[2] = camera->savebuf[i]; 
	  fwrite (buff, 1, 3, outfile); 
	}
      }
      else {
	for (i = 0; i < camera->vid_win.width * camera->vid_win.height * 3; i++) {
	  fputc(camera->savebuf[i], outfile);
	}
      }
      
    } 
    else if (GTK_TOGGLE_BUTTON(page_ppm.radio_ascii_ppm)->active){
      fprintf (outfile, "P3\n%d %d\n%d\n", camera->vid_win.width, camera->vid_win.height, 255);
      if (camera->greyscale) {
	for (i = 0; i < camera->vid_win.width*camera->vid_win.height; i++){ 
	  fprintf (outfile, "%03d %03d %03d\n", camera->savebuf[i], camera->savebuf[i], camera->savebuf[i]);
	}
      }
      else
	for (i = 0; i < camera->vid_win.width*camera->vid_win.height*3; i=i+3){ 
	  fprintf (outfile, "%03d %03d %03d\n", camera->savebuf[i], camera->savebuf[i+1], camera->savebuf[i+2]);
	}
    }
    
    break; 
    
  case JPEG:
    cinfo.err = jpeg_std_error(&jerr);
    jpeg_create_compress(&cinfo);
    
    jpeg_stdio_dest(&cinfo, outfile);
    cinfo.image_width = camera->vid_win.width;
    cinfo.image_height = camera->vid_win.height;
            
    if (camera->greyscale) {
      cinfo.input_components = 1;   
      cinfo.in_color_space = JCS_GRAYSCALE;     
    }
    else {
      cinfo.input_components = 3;
      cinfo.in_color_space = JCS_RGB;
    }
    
    cinfo.smoothing_factor = GTK_ADJUSTMENT(page_jpeg.smooth_adj_jpeg)->value;
    if (GTK_TOGGLE_BUTTON(page_jpeg.optimize_jpeg)->active) {
      cinfo.optimize_coding = TRUE;
    }
    
    jpeg_set_defaults(&cinfo);
    jpeg_set_quality(&cinfo, GTK_ADJUSTMENT(page_jpeg.quality_adj_jpeg)->value, TRUE);
    jpeg_start_compress(&cinfo, TRUE);
    row_stride = cinfo.input_components*camera->vid_win.width;
    while (cinfo.next_scanline < cinfo.image_height) {
      row_pointer[0] = &camera->savebuf[cinfo.next_scanline * row_stride];
      jpeg_write_scanlines(&cinfo, row_pointer, 1);
    }
    jpeg_finish_compress(&cinfo);
    jpeg_destroy_compress(&cinfo);

    break;
  case gif:
    

    break;
  }

  fclose(outfile);
  if(!camera->frozen)
    free(camera->savebuf);
  camera->saving = 0;
  
  gtk_widget_destroy(widget->parent->parent->parent);
  return;
}

void save_cancel(GtkWidget *widget, struct Camera *camera)
{
  pthread_mutex_lock(&camera->pic_mutex);
  gtk_widget_destroy(widget->parent->parent->parent);
  if(!camera->frozen)
    free(camera->savebuf);
  camera->saving = 0;
  pthread_mutex_unlock(&camera->pic_mutex);
  return;
}

void destroy( GtkWidget *widget,
	      gpointer   data )
{
  gtk_main_quit();
}

void switch_page(GtkNotebook *notebook, gpointer page, guint page_num, struct Camera *camera)
{
  camera->savetype = page_num;
  
}

void save_dialog(GtkWidget *widget, struct Camera *camera)
{
  GtkWidget *notebook;
  GtkWidget *window;
  GtkWidget *label;
  GtkWidget *button;
  GtkWidget *vbox, *hbox;
  GtkObject *quality_adj_jpeg;
  int openpage = camera->savetype;

  pthread_mutex_lock(&camera->pic_mutex);

  if(camera->greyscale){
    camera->savebuf = malloc(camera->vid_caps.maxwidth*camera->vid_caps.maxheight);
    memcpy(camera->savebuf, camera->pic, camera->vid_caps.maxwidth*camera->vid_caps.maxheight);
  }
  else{
    camera->savebuf = malloc(camera->vid_caps.maxwidth*camera->vid_caps.maxheight*3);
    memcpy(camera->savebuf, camera->pic, camera->vid_caps.maxwidth*camera->vid_caps.maxheight*3);
  }

  camera->saving = 1;

  pthread_mutex_unlock(&camera->pic_mutex);

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (window), "Gqcam: Save Image");
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_add (GTK_CONTAINER (window), vbox);
  gtk_widget_show(vbox);
  
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 2);
  label = gtk_label_new("File");
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 2);
  gtk_widget_show(label);
  textbox = gtk_entry_new();
  gtk_box_pack_start(GTK_BOX(hbox), textbox, FALSE, FALSE, 2);
  gtk_widget_show(textbox);
  button = gtk_button_new_with_label("Browse...");
  gtk_signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC (file_select), NULL);
  gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 2);
  gtk_widget_show(button);
  gtk_widget_show(hbox);

  notebook = gtk_notebook_new();
  gtk_box_pack_start(GTK_BOX(vbox), notebook, FALSE, FALSE, 2);
  gtk_widget_show(notebook);
  gtk_signal_connect (GTK_OBJECT (notebook), "switch-page",(GtkSignalFunc) switch_page, camera);
  
  /* PPM PAGE */
  page_ppm.frame_ppm = gtk_frame_new("Save As");
  gtk_frame_set_shadow_type(GTK_FRAME(page_ppm.frame_ppm), GTK_SHADOW_NONE);
  gtk_widget_show(page_ppm.frame_ppm);

  page_ppm.vbox_ppm = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(page_ppm.frame_ppm), page_ppm.vbox_ppm);
  gtk_widget_show(page_ppm.vbox_ppm);

  page_ppm.radio_raw_ppm = gtk_radio_button_new_with_label(NULL, "Raw");
  gtk_box_pack_start(GTK_BOX(page_ppm.vbox_ppm), page_ppm.radio_raw_ppm, FALSE, FALSE, 2);
  gtk_widget_show(page_ppm.radio_raw_ppm);
  
  page_ppm.group_ppm = gtk_radio_button_group(GTK_RADIO_BUTTON(page_ppm.radio_raw_ppm));
  page_ppm.radio_ascii_ppm = gtk_radio_button_new_with_label(page_ppm.group_ppm, "ASCII");
  gtk_box_pack_start(GTK_BOX(page_ppm.vbox_ppm), page_ppm.radio_ascii_ppm, FALSE, FALSE, 2);
  gtk_widget_show(page_ppm.radio_ascii_ppm);

  page_ppm.label_ppm = gtk_label_new("PPM");
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), page_ppm.frame_ppm, page_ppm.label_ppm);

  /* JPEG PAGE */
  page_jpeg.frame_jpeg = gtk_frame_new(NULL);
  gtk_frame_set_shadow_type(GTK_FRAME(page_jpeg.frame_jpeg), GTK_SHADOW_NONE);
  gtk_widget_show(page_jpeg.frame_jpeg);

  page_jpeg.vbox_jpeg = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(page_jpeg.frame_jpeg), page_jpeg.vbox_jpeg);
  gtk_widget_show(page_jpeg.vbox_jpeg);

  page_jpeg.table_jpeg = gtk_table_new(3, 3, TRUE);
  gtk_box_pack_start(GTK_BOX(page_jpeg.vbox_jpeg), page_jpeg.table_jpeg, FALSE, FALSE, 2);
  gtk_widget_show(page_jpeg.table_jpeg);

  page_jpeg.label_jpeg = gtk_label_new("Quality");
  gtk_table_attach_defaults(GTK_TABLE(page_jpeg.table_jpeg), page_jpeg.label_jpeg, 0, 1, 0, 1);
  gtk_widget_show(page_jpeg.label_jpeg);

  page_jpeg.quality_adj_jpeg = gtk_adjustment_new(75, 0, 100, 1, 10, 1);
  page_jpeg.quality_jpeg = gtk_hscale_new(GTK_ADJUSTMENT(page_jpeg.quality_adj_jpeg));
  gtk_scale_set_digits (GTK_SCALE(page_jpeg.quality_jpeg), 0);
  gtk_table_attach_defaults(GTK_TABLE(page_jpeg.table_jpeg), page_jpeg.quality_jpeg, 1, 3, 0, 1);
  gtk_widget_show(page_jpeg.quality_jpeg);

  page_jpeg.label_jpeg = gtk_label_new("Smoothness");
  gtk_table_attach_defaults(GTK_TABLE(page_jpeg.table_jpeg), page_jpeg.label_jpeg, 0, 1, 1, 2);
  gtk_widget_show(page_jpeg.label_jpeg);

  page_jpeg.smooth_adj_jpeg = gtk_adjustment_new( 0, 0, 100, 1, 10, 1 );
  page_jpeg.smooth_jpeg = gtk_hscale_new(GTK_ADJUSTMENT(page_jpeg.smooth_adj_jpeg));
  gtk_range_set_update_policy (GTK_RANGE(page_jpeg.smooth_jpeg), GTK_UPDATE_CONTINUOUS);
  gtk_scale_set_digits (GTK_SCALE(page_jpeg.smooth_jpeg), 0);
  gtk_scale_set_value_pos (GTK_SCALE(page_jpeg.smooth_jpeg), GTK_POS_TOP);
  gtk_scale_set_draw_value (GTK_SCALE(page_jpeg.smooth_jpeg), TRUE);
  gtk_table_attach_defaults(GTK_TABLE(page_jpeg.table_jpeg), page_jpeg.smooth_jpeg, 1, 3, 1, 2);
  gtk_widget_show(page_jpeg.smooth_jpeg);

  page_jpeg.optimize_jpeg = gtk_check_button_new_with_label("Optimize");
  gtk_table_attach_defaults(GTK_TABLE(page_jpeg.table_jpeg), page_jpeg.optimize_jpeg, 0, 1, 2, 3);
  gtk_widget_show(page_jpeg.optimize_jpeg);

  page_jpeg.label_jpeg = gtk_label_new("Jpeg");
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), page_jpeg.frame_jpeg, page_jpeg.label_jpeg);

/* GIF PAGE */
/*
  page_gif.frame_gif = gtk_frame_new(NULL);
  gtk_frame_set_shadow_type(GTK_FRAME(page_gif.frame_gif), GTK_SHADOW_NONE);
  gtk_widget_show(page_gif.frame_gif);

  page_gif.vbox_gif = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(page_gif.frame_gif), page_gif.vbox_gif);
  gtk_widget_show(page_gif.vbox_gif);

  page_gif.interlace_gif = gtk_check_button_new_with_label("Interlace");
  gtk_box_pack_start(GTK_BOX(page_gif.vbox_gif), page_gif.interlace_gif, FALSE, FALSE, 2);
  gtk_widget_show(page_gif.interlace_gif);

  page_gif.comment_gif = gtk_check_button_new_with_label("Comment:");
  gtk_box_pack_start(GTK_BOX(page_gif.vbox_gif), page_gif.comment_gif, FALSE, FALSE, 2);
  gtk_widget_show(page_gif.comment_gif);

  page_gif.comment_text_gif = gtk_text_new( NULL, NULL);
  gtk_box_pack_start(GTK_BOX(page_gif.vbox_gif), page_gif.comment_text_gif, FALSE, FALSE, 2);
  gtk_text_set_editable(GTK_TEXT(page_gif.comment_text_gif), TRUE);
  gtk_widget_set_usize(GTK_WIDGET(page_gif.comment_text_gif), 100, 50);
  gtk_widget_show(page_gif.comment_text_gif);
  
  page_gif.label_gif = gtk_label_new("Gif");
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), page_gif.frame_gif, page_gif.label_gif);
*/

/* END PAGES */
  gtk_notebook_set_page(GTK_NOTEBOOK(notebook), camera->savetype);
  hbox = gtk_hbox_new(TRUE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 2);
  gtk_widget_show(hbox);
  button = gtk_button_new_with_label("Ok");
  gtk_signal_connect(GTK_OBJECT(button), "clicked", save_ok, camera);
  gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, TRUE, 5);
  gtk_widget_show(button);
  button = gtk_button_new_with_label("Cancel");
  gtk_signal_connect(GTK_OBJECT(button), "clicked", save_cancel, camera);
  gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, TRUE, 5);
  gtk_widget_show(button);
  
  gtk_widget_show(window);
  return;
}




